#include <glib.h>
#include <SDL.h>
#include "racanhack.h"
#include "cave.h"

static struct _cave *cave;
static void line_map_pass(int x0, int y0, int x1, int y1);
static void line_waynavi(int x0, int y0, int x1, int y1);
static void cave_rect_split(struct _rect *rect_parent);
static void cave_junction_more();
static void cave_room_make();
static void cave_map_set();
static void cave_map_print_debug();
static struct _rect *cave_rect_add(int lx, int ly, int hx, int hy);
static struct _junction *cave_junction_add(struct _rect *rect_u, struct _rect *rect_d, struct _rect *rect_l, struct _rect *rect_r);
static struct _room *cave_room_add(int lx, int ly, int hx, int hy, gboolean route);

static void line_map_pass(int x0, int y0, int x1, int y1)
{
  int min_x, max_x, min_y, max_y, i, j;
  min_x = MIN(x0, x1);
  max_x = MAX(x0, x1);
  min_y = MIN(y0, y1);
  max_y = MAX(y0, y1);
  g_assert((min_x >= 0) && (max_x < MAP_W) && (min_y >= 0) && (max_y < MAP_H));
  if ((x0 <= x1) && (y0 >= y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][max_y].pass = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[max_x][j].pass = TRUE;
    return;
  };
  if ((x0 > x1) && (y0 > y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][min_y].pass = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[max_x][j].pass = TRUE;
    return;
  };
  if ((x0 > x1) && (y0 <= y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][min_y].pass = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[min_x][j].pass = TRUE;
    return;
  };
  if ((x0 <= x1) && (y0 < y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][max_y].pass = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[min_x][j].pass = TRUE;
    return;
  };
}

static void line_waynavi(int x0, int y0, int x1, int y1)
{
  int min_x, max_x, min_y, max_y, i, j;
  min_x = MIN(x0, x1);
  max_x = MAX(x0, x1);
  min_y = MIN(y0, y1);
  max_y = MAX(y0, y1);
  g_assert((min_x >= 0) && (max_x < MAP_W) && (min_y >= 0) && (max_y < MAP_H));
  if ((x0 <= x1) && (y0 >= y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][max_y].waynavi = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[max_x][j].waynavi = TRUE;
    return;
  };
  if ((x0 > x1) && (y0 > y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][min_y].waynavi = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[max_x][j].waynavi = TRUE;
    return;
  };
  if ((x0 > x1) && (y0 <= y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][min_y].waynavi = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[min_x][j].waynavi = TRUE;
    return;
  };
  if ((x0 <= x1) && (y0 < y1)) {
    for (i = min_x; i <= max_x; i++) cave->map[i][max_y].waynavi = TRUE;
    for (j = min_y; j <= max_y; j++) cave->map[min_x][j].waynavi = TRUE;
    return;
  };
}

void cave_generate(struct _cave *c)
{
  int i, j;
  GList *li, *next;
  struct _room *room;
  cave = c;
  for (i = 0; i < MAP_W; i++) {
    for (j = 0; j < MAP_H; j++) {
      cave->map[i][j].pass = FALSE;
      cave->map[i][j].waynavi = FALSE;
      cave->map[i][j].passed = FALSE;
      cave->map[i][j].disturb = FALSE;
      cave->map[i][j].room = NULL;
    };
  };
  cave->rect_header = NULL;
  cave->junction_header = NULL;
  cave->room_header = NULL;
  cave_rect_split(cave_rect_add(4, 4, MAP_W - 5, MAP_H - 5));
  cave_junction_more();
  cave_room_make();
  cave_map_set();
  cave_map_print_debug();
  for (li = g_list_first(cave->room_header); li != NULL; li = next) {
    next = g_list_next(li);
    room = (struct _room *)li->data;
    if (room->route == TRUE) {
      cave->room_header = g_list_remove(cave->room_header, room);
      my_free(room);
    };
  };
}

void cave_generate_tutorial(struct _cave *c)
{
  int i, j;
  struct _room *room;
  cave = c;
  for (i = 0; i < MAP_W; i++) {
    for (j = 0; j < MAP_H; j++) {
      cave->map[i][j].pass = FALSE;
      cave->map[i][j].waynavi = FALSE;
      cave->map[i][j].passed = FALSE;
      cave->map[i][j].disturb = FALSE;
      cave->map[i][j].room = NULL;
    };
  };
  cave->rect_header = NULL;
  cave->junction_header = NULL;
  cave->room_header = NULL;
  room = cave_room_add(2, 2, 10, 10, FALSE);
  for (i = room->lx; i <= room->hx; i++) {
    for (j = room->ly; j <= room->hy; j++) {
      cave->map[i][j].pass = TRUE;
      cave->map[i][j].room = room;
    };
  };
}

void cave_free(struct _cave *c)
{
  g_list_foreach(c->rect_header, (GFunc)my_free, NULL);
  g_list_free(c->rect_header);
  g_list_foreach(c->room_header, (GFunc)my_free, NULL);
  g_list_free(c->room_header);
  g_list_foreach(c->junction_header, (GFunc)my_free, NULL);
  g_list_free(c->junction_header);
}

static void cave_rect_split(struct _rect *rect_parent)
{
  struct _rect *rect_child;
  if (rect_parent->hy - rect_parent->ly < 16) {
    rect_parent->u = TRUE;
    rect_parent->d = TRUE;
  };
  if (rect_parent->hx - rect_parent->lx < 16) {
    rect_parent->l = TRUE;
    rect_parent->r = TRUE;
  };
  if (g_random_int_range(0, 8) == 0) {
    rect_parent->u = TRUE;
    rect_parent->d = TRUE;
    rect_parent->l = TRUE;
    rect_parent->r = TRUE;
  };
  if ((rect_parent->u == TRUE) &&
      (rect_parent->d == TRUE) &&
      (rect_parent->l == TRUE) &&
      (rect_parent->r == TRUE)) return;
  rect_child = cave_rect_add(rect_parent->lx, rect_parent->ly,
			     rect_parent->hx, rect_parent->hy);
  if (rect_parent->u == FALSE) {
    rect_parent->ly = g_random_int_range(rect_parent->ly + 8, rect_parent->hy - 7);
    rect_child->hy = rect_parent->ly;
    rect_parent->u = TRUE;
    rect_child->d = TRUE;
    cave_junction_add(rect_child, rect_parent, NULL, NULL);
    cave_rect_split(rect_parent);
    cave_rect_split(rect_child);
    return;
  };
  if (rect_parent->d == FALSE) {
    rect_parent->hy = g_random_int_range(rect_parent->ly + 8, rect_parent->hy - 7);
    rect_child->ly = rect_parent->hy;
    rect_parent->d = TRUE;
    rect_child->u = TRUE;
    cave_junction_add(rect_parent, rect_child, NULL, NULL);
    cave_rect_split(rect_parent);
    cave_rect_split(rect_child);
    return;
  };
  if (rect_parent->l == FALSE) {
    rect_parent->lx = g_random_int_range(rect_parent->lx + 8, rect_parent->hx - 7);
    rect_child->hx = rect_parent->lx;
    rect_parent->l = TRUE;
    rect_child->r = TRUE;
    cave_junction_add(NULL, NULL, rect_child, rect_parent);
    cave_rect_split(rect_parent);
    cave_rect_split(rect_child);
    return;
  };
  if (rect_parent->r == FALSE) {
    rect_parent->hx = g_random_int_range(rect_parent->lx + 8, rect_parent->hx - 7);
    rect_child->lx = rect_parent->hx;
    rect_parent->r = TRUE;
    rect_child->l = TRUE;
    cave_junction_add(NULL, NULL, rect_parent, rect_child);
    cave_rect_split(rect_parent);
    cave_rect_split(rect_child);
    return;
  };
}

static void cave_junction_more()
{
  GList *li;
  struct _rect *rect;
  struct _rect *rectmap[MAP_W][MAP_H];
  int i, j;
  for (li = g_list_first(cave->rect_header); li != NULL; li = g_list_next(li)) {
    rect = (struct _rect *)li->data;
    for (i = rect->lx; i <= rect->hx; i++) {
      for (j = rect->ly; j <= rect->hy; j++) {
      	rectmap[i][j] = rect;
      };
    };
  };
  for (i = 4; i < MAP_W - 4; i++) {
    for (j = 4; j < MAP_H - 4 - 1; j++) {
      if (rectmap[i][j] != rectmap[i][j + 1]) {
	if (g_random_int_range(0, 63) == 0) {
	  cave_junction_add(rectmap[i][j], rectmap[i][j + 1], NULL, NULL);
	};
      };
    };
  };
}

static void cave_room_make()
{
  GList *li;
  struct _rect *rect;
  int x, y, w, h;
  gboolean route;
  for (li = g_list_first(cave->rect_header); li != NULL; li = g_list_next(li)) {
    rect = (struct _rect *)li->data;
    w = g_random_int_range(3, rect->hx - rect->lx - 3);
    h = g_random_int_range(3, rect->hy - rect->ly - 3);
    x = g_random_int_range(rect->lx + 2, rect->hx - 1 - w);
    y = g_random_int_range(rect->ly + 2, rect->hy - 1 - h);
    route = FALSE;
    if (g_random_int_range(0,20) == 0) {
      if (g_list_length(cave->rect_header) > 1) {
	route = TRUE;
      };
    };
    rect->room = cave_room_add(x, y, x + w, y + h, route);
  };
}

static void cave_map_set()
{
  GList *li;
  struct _room *room;
  int i, j;
  struct _junction *junction;
  for (li = g_list_first(cave->room_header); li != NULL; li = g_list_next(li)) {
    room = (struct _room *)li->data;
    if (room->route == TRUE) continue;
    for (i = room->lx; i <= room->hx; i++) {
      for (j = room->ly; j <= room->hy; j++) {
      	cave->map[i][j].pass = TRUE;
      	cave->map[i][j].room = room;
      };
    };
  };
  for (li = g_list_first(cave->junction_header); li != NULL; li = g_list_next(li)) {
    junction = (struct _junction *)li->data;
    if ((junction->rect_l != NULL) && (junction->rect_r != NULL)) {
      g_assert((junction->rect_u == NULL) && (junction->rect_d == NULL));
      g_assert(junction->rect_l->hx == junction->rect_r->lx);
      junction->couner0.x = junction->rect_l->hx;
      junction->couner0.y = g_random_int_range(junction->rect_l->room->ly + 1, junction->rect_l->room->hy);
      junction->couner1.x = junction->rect_r->lx;
      junction->couner1.y = g_random_int_range(junction->rect_r->room->ly + 1, junction->rect_r->room->hy);
      cave->map[junction->couner0.x][junction->couner0.y].disturb = TRUE;
      cave->map[junction->couner1.x][junction->couner1.y].disturb = TRUE;
      cave->map[junction->rect_l->room->hx][junction->couner0.y].disturb = TRUE;
      cave->map[junction->rect_l->room->hx + 1][junction->couner0.y].disturb = TRUE;
      cave->map[junction->rect_r->room->lx][junction->couner1.y].disturb = TRUE;
      cave->map[junction->rect_r->room->lx - 1][junction->couner1.y].disturb = TRUE;
      line_map_pass(junction->couner0.x, junction->couner0.y,
 		   junction->couner1.x, junction->couner1.y);
      line_map_pass(junction->rect_l->room->hx, junction->couner0.y,
		   junction->couner0.x, junction->couner0.y);
      line_map_pass(junction->rect_r->room->lx, junction->couner1.y,
		   junction->couner1.x, junction->couner1.y);
      if (junction->rect_l->room->route == TRUE) {
	line_map_pass(junction->rect_l->room->hx, junction->couner0.y,
		     (junction->rect_l->room->lx + junction->rect_l->room->hx) / 2,
		     (junction->rect_l->room->ly + junction->rect_l->room->hy) / 2);
      };
      if (junction->rect_r->room->route == TRUE) {
	line_map_pass(junction->rect_r->room->lx, junction->couner1.y,
		     (junction->rect_r->room->lx + junction->rect_r->room->hx) / 2,
		     (junction->rect_r->room->ly + junction->rect_r->room->hy) / 2);
      };
      line_waynavi(junction->couner0.x, junction->couner0.y,
			junction->couner1.x, junction->couner1.y);
      line_waynavi(junction->rect_l->room->hx, junction->couner0.y,
			junction->couner0.x, junction->couner0.y);
      line_waynavi(junction->rect_r->room->lx, junction->couner1.y,
			junction->couner1.x, junction->couner1.y);
      line_waynavi(junction->rect_l->room->hx, junction->couner0.y,
			(junction->rect_l->room->lx + junction->rect_l->room->hx) / 2,
			(junction->rect_l->room->ly + junction->rect_l->room->hy) / 2);
      line_waynavi(junction->rect_r->room->lx, junction->couner1.y,
			(junction->rect_r->room->lx + junction->rect_r->room->hx) / 2,
			(junction->rect_r->room->ly + junction->rect_r->room->hy) / 2);
      continue;
    };
    if ((junction->rect_u != NULL) && (junction->rect_d != NULL)) {
      g_assert((junction->rect_l == NULL) && (junction->rect_r == NULL));
      g_assert(junction->rect_u->hy == junction->rect_d->ly);
      junction->couner0.x = g_random_int_range(junction->rect_u->room->lx + 1, junction->rect_u->room->hx);
      junction->couner0.y = junction->rect_u->hy;
      junction->couner1.x = g_random_int_range(junction->rect_d->room->lx + 1, junction->rect_d->room->hx);
      junction->couner1.y = junction->rect_d->ly;
      cave->map[junction->couner0.x][junction->couner0.y].disturb = TRUE;
      cave->map[junction->couner1.x][junction->couner1.y].disturb = TRUE;
      cave->map[junction->couner0.x][junction->rect_u->room->hy].disturb = TRUE;
      cave->map[junction->couner0.x][junction->rect_u->room->hy + 1].disturb = TRUE;
      cave->map[junction->couner1.x][junction->rect_d->room->ly].disturb = TRUE;
      cave->map[junction->couner1.x][junction->rect_d->room->ly - 1].disturb = TRUE;
      line_map_pass(junction->couner0.x, junction->couner0.y,
		   junction->couner1.x, junction->couner1.y);
      line_map_pass(junction->couner0.x, junction->rect_u->room->hy,
		   junction->couner0.x, junction->couner0.y);
      line_map_pass(junction->couner1.x, junction->rect_d->room->ly,
		   junction->couner1.x, junction->couner1.y);
      if (junction->rect_u->room->route == TRUE) {
	line_map_pass(junction->couner0.x, junction->rect_u->room->hy,
		      (junction->rect_u->room->lx + junction->rect_u->room->hx) / 2,
		      (junction->rect_u->room->ly + junction->rect_u->room->hy) / 2);
      };
      if (junction->rect_d->room->route == TRUE) {
	line_map_pass(junction->couner1.x, junction->rect_d->room->ly,
		     (junction->rect_d->room->lx + junction->rect_d->room->hx) / 2,
		     (junction->rect_d->room->ly + junction->rect_d->room->hy) / 2);
      };
      line_waynavi(junction->couner0.x, junction->couner0.y,
			junction->couner1.x, junction->couner1.y);
      line_waynavi(junction->couner0.x, junction->rect_u->room->hy,
			junction->couner0.x, junction->couner0.y);
      line_waynavi(junction->couner1.x, junction->rect_d->room->ly,
			junction->couner1.x, junction->couner1.y);
      line_waynavi(junction->couner0.x, junction->rect_u->room->hy,
			(junction->rect_u->room->lx + junction->rect_u->room->hx) / 2,
			(junction->rect_u->room->ly + junction->rect_u->room->hy) / 2);
      line_waynavi(junction->couner1.x, junction->rect_d->room->ly,
			(junction->rect_d->room->lx + junction->rect_d->room->hx) / 2,
			(junction->rect_d->room->ly + junction->rect_d->room->hy) / 2);
      continue;
    };
  };
}

static void cave_map_print_debug()
{
  int i, j;
  int debug_map[MAP_W][MAP_H];
  GList *li;
  struct _rect *rect;
  return;
  for (i = 0; i < MAP_W; i++) {
    for (j = 0; j < MAP_H; j++) {
      debug_map[i][j] = 0;
    };
  };
  for (li = g_list_first(cave->rect_header); li != NULL; li = g_list_next(li)) {
    rect = (struct _rect *)li->data;
    for (i = rect->lx, j = rect->ly; i <= rect->hx; i++) { 
      debug_map[i][j] = 1; 
    }; 
    for (i = rect->lx, j = rect->hy; i <= rect->hx; i++)  { 
      debug_map[i][j] = 1;  
    };   
    for (i = rect->lx, j = rect->ly; j <= rect->hy; j++)  {   
      debug_map[i][j] = 1;  
    };   
    for (i = rect->hx, j = rect->ly; j <= rect->hy; j++)  {   
      debug_map[i][j] = 1;
    };   
  };
  for (i = 0; i < MAP_W; i++) {
    for (j = 0; j < MAP_H; j++) {
      if (cave->map[i][j].pass == TRUE) {
	debug_map[i][j] = 2;
	if (cave->map[i][j].waynavi == TRUE) {
	  debug_map[i][j] = 3;
	};
	if (cave->map[i][j].disturb == TRUE) {
	  debug_map[i][j] = 4;
	};
      };
    };
  };
  for (j = 0; j < MAP_H; j++) { 
    for (i = 0; i < MAP_W; i++) { 
      switch (debug_map[i][j]) { 
      case 0:
	/* MAP_WALL */
	g_print(" "); 
	break; 
      case 1:
	/* RECT'S BORDER */
	g_print("."); 
	break; 
      case 2:
	/* MAP_PASS */ 
	g_print("#"); 
	break; 
      case 3:
	/* MAP_PASS AND WAYNAVI(monster's route) */
	g_print("+"); 
	break; 
      case 4:
	/* MAP_PASS AND DISTURB */
	g_print("*");
	break;
      };
    }; 
    g_print("\n"); 
  }; 
}

static struct _rect *cave_rect_add(int lx, int ly, int hx, int hy)
{
  struct _rect *rect;
  rect = my_new(struct _rect, 1);
  rect->lx = lx;
  rect->ly = ly;
  rect->hx = hx;
  rect->hy = hy;
  rect->u = FALSE;
  rect->d = FALSE;
  rect->l = FALSE;
  rect->r = FALSE;
  cave->rect_header = g_list_append(cave->rect_header, rect);
  return(rect);
}

static struct _junction *cave_junction_add(struct _rect *rect_u, struct _rect *rect_d, struct _rect *rect_l, struct _rect *rect_r)
{
  struct _junction *junction;
  junction = my_new(struct _junction, 1);
  junction->rect_u = rect_u;
  junction->rect_d = rect_d;
  junction->rect_l = rect_l;
  junction->rect_r = rect_r;
  cave->junction_header = g_list_append(cave->junction_header, junction);
  return(junction);
}

static struct _room *cave_room_add(int lx, int ly, int hx, int hy, gboolean route)
{
  struct _room *room;
  room = my_new(struct _room, 1);
  room->lx = lx;
  room->ly = ly;
  room->hx = hx;
  room->hy = hy;
  room->route = route;
  room->monsterhouse = FALSE;
  cave->room_header = g_list_append(cave->room_header, room);
  return(room);
}
