#include <stdlib.h>
#include <glib.h>
#include <SDL.h>
#include "racanhack.h"
#include "task.h"
#include "video.h"
#include "cave.h"
#include "player.h"
#include "monster2.h"
#include "object.h"
#include "object2.h"
#include "floor.h"

static int object_calc_id(struct _object *object);
static int object_calc_plus(struct _object *object);
static int object_calc_count(struct _object *object);
static void init_object_info(int id, int type, gchar *ident_name, gchar *unident_name, int atk, int ac);
static int randomize_order(const void *a0, const void *a1);
static void weapon_init();
static void shield_init();
static void arrow_init();
static void potion_init();
static void scroll_init();
static void food_init();
static void extra_init();

int damage_calc_wielded(struct _monster *monster, struct _object *object, struct _player *player)
{
  int d;
  int i;
  int ac;
  if (g_random_int_range(0, 8) == 0) return DODGED_DAMAGE;
  d = player->atk + ((player->atk * (object_atk(object) + player->cstr - 8)) / 16);
  d *= g_random_int_range(112, 143);
  ac = monster_ac(monster);
  for (i = 0; i < ac; i++) {
    d -= (d / 16);
  };
  d /= 128;
  return d;
}

int damage_calc_throwed(struct _monster *monster, struct _object *object, struct _player *player)
{
  int d;
  int i;
  int ac;
  if (g_random_int_range(0, 8) == 0) return DODGED_DAMAGE;
  d = player->atk + ((player->atk * (object_atk(object) - 8)) / 16);
  d *= g_random_int_range(112, 143);
  ac = monster_ac(monster);
  for (i = 0; i < ac; i++) {
    d -= (d / 16);
  };
  d /= 128;
  return d;
}

int damage_calc_attacked_by_monster(struct _monster *monster, struct _object *object, struct _player *player)
{
  int d;
  int i;
  int ac;
  if (g_random_int_range(0, 8) == 0) return DODGED_DAMAGE;
  d = monster_atk(monster);
  d *= g_random_int_range(112, 143);
  ac = object_ac(object);
  for (i = 0; i < ac; i++) {
    d -= (d / 16);
  };
  d /= 128;
  if (d == 0) d = 1;
  return d;
}

gboolean can_detach(struct _object *object, struct _player *player)
{
  if (object == player->weapon) {
    if (object_cursed(player->weapon)) {
      return FALSE;
    };
  };
  if (object == player->shield) {
    if (object_cursed(player->shield)) {
      return FALSE;
    };
  };
  return TRUE;

}

static int object_calc_id(struct _object *object)
{
  int r;
  int type;
  r = g_random_int_range(0, 100);
  if (r < 15) {
    type = OBJECT_TYPE_WEAPON;
  } else if (r < 30) {
    type = OBJECT_TYPE_SHIELD;
  } else if (r < 45) {
    type = OBJECT_TYPE_ARROW;
  } else if (r < 70) {
    type = OBJECT_TYPE_POTION;
  } else if (r < 95) {
    type = OBJECT_TYPE_SCROLL;
  } else if (r < 100) {
    type = OBJECT_TYPE_FOOD;
  } else {
    g_assert_not_reached();
  };
  r = g_random_int_range(0, 100);
  switch (type) {
  case OBJECT_TYPE_WEAPON:
    if (r < 45) return OBJECT_ID_DAGGER;
    if (r < 70) return OBJECT_ID_MACE;
    if (r < 90) return OBJECT_ID_LONGSWORD;
    if (r < 100) return OBJECT_ID_TWOHSWORD;
  case OBJECT_TYPE_SHIELD:
    if (r < 35) return OBJECT_ID_SMALL_SHIELD;
    if (r < 55) return OBJECT_ID_ORCISH_SHIELD;
    if (r < 75) return OBJECT_ID_ELVEN_SHIELD;
    if (r < 90) return OBJECT_ID_LARGE_SHIELD;
    if (r < 100) return OBJECT_ID_DWARVISH_SHIELD;
  case OBJECT_TYPE_ARROW:
    if (r < 75) return OBJECT_ID_WOOD_ARROW;
    if (r < 100) return OBJECT_ID_IRON_ARROW;
  case OBJECT_TYPE_POTION:
    if (r < 40) return OBJECT_ID_HEAL_POTION;
    if (r < 55) return OBJECT_ID_POISON_POTION;
    if (r < 80) return OBJECT_ID_RESSTR_POTION;
    if (r < 95) return OBJECT_ID_INCSTR_POTION;
    if (r < 100) return OBJECT_ID_RAISELEV_POTION;
  case OBJECT_TYPE_SCROLL:
    if (r < 20) return OBJECT_ID_IDENT_SCROLL;
    if (r < 25) return OBJECT_ID_RMCURSE_SCROLL;
    if (r < 45) return OBJECT_ID_ENCWEAPON_SCROLL;
    if (r < 60) return OBJECT_ID_ENCSHIELD_SCROLL;
    if (r < 75) return OBJECT_ID_EXPLOSION_SCROLL;
    if (r < 80) return OBJECT_ID_MONDETECT_SCROLL;
    if (r < 90) return OBJECT_ID_OBJDETECT_SCROLL;
    if (r < 100) return OBJECT_ID_MAP_SCROLL;
  case OBJECT_TYPE_FOOD:
    if (r < 70) return OBJECT_ID_SLIMEMOLD;
    if (r < 100) return OBJECT_ID_RATION;
  };
  g_assert_not_reached();
  return OBJECT_ID_DAGGER;
}

static int object_calc_plus(struct _object *object)
{
  int p;
  if ((object_type(object) == OBJECT_TYPE_WEAPON) ||
      (object_type(object) == OBJECT_TYPE_SHIELD)) {
    p = g_random_int_range(0,100);
    if (p < 2) return -2;
    if (p < 5) return -1;
    if (p < 94) return 0;
    if (p < 98) return 1;
    if (p < 100) return 2;
    g_assert_not_reached();
  };
  return 0;
};

static int object_calc_count(struct _object *object)
{
  if (object_type(object) == OBJECT_TYPE_ARROW) return g_random_int_range(1, 8);
  return 1;
};

struct _object *object_add()
{
  struct _object *object;
  object = my_new(struct _object, 1);
  object->id = object_calc_id(object);
  object->coord = rand_coord_on_room();
  object->plus = object_calc_plus(object);
  object->count = object_calc_count(object);
  object->sprite = NULL;
  object->task = NULL;
  curf->object_header = g_list_append(curf->object_header, object);
  return (object);
}

struct _object *object_add_monsterhouse(struct _coord c)
{
  struct _object *object;
  object = my_new(struct _object, 1);
  object->id = object_calc_id(object);
  object->coord = c;
  object->plus = object_calc_plus(object);
  object->count = object_calc_count(object);
  object->sprite = NULL;
  object->task = NULL;
  curf->object_header = g_list_append(curf->object_header, object);
  return (object);
}

struct _object *noweapon_add()
{
  struct _object *noweapon;
  noweapon = &Noweapon;
  noweapon->id = OBJECT_ID_NOWEAPON;
  noweapon->plus = 0;
  noweapon->count = 0;
  noweapon->sprite = NULL;
  noweapon->task = NULL;
  return (noweapon);
}

struct _object *noshield_add()
{
  struct _object *noshield;
  noshield = &Noshield;
  noshield->id = OBJECT_ID_NOSHIELD;
  noshield->plus = 0;
  noshield->count = 0;
  noshield->sprite = NULL;
  noshield->task = NULL;
  return (noshield);
}

struct _object *noarrow_add()
{
  struct _object *noarrow;
  noarrow = &Noarrow;
  noarrow->id = OBJECT_ID_NOARROW;
  noarrow->plus = 0;
  noarrow->count = 0;
  noarrow->sprite = NULL;
  noarrow->task = NULL;
  return (noarrow);
}

struct _object *object_add_specify_id(int id)
{
  struct _object *object;
  object = my_new(struct _object, 1);
  object->id = id;
  object->coord = rand_coord_on_room();
  object->plus = object_calc_plus(object);
  object->count = object_calc_count(object);
  object->sprite = video_sprite_add(OBJECT_SPRITE_Z_ORDER);
  object->task = NULL;
  curf->object_header = g_list_append(curf->object_header, object);
  return (object);
}

struct _object *arrow_add_from_quiver(struct _object *arrow_src)
{
  struct _object *arrow;
  arrow = my_new(struct _object, 1);
  arrow->id = arrow_src->id;
  arrow->coord = arrow_src->coord;
  arrow->plus = arrow_src->plus;
  arrow->count = 1;
  arrow->sprite = video_sprite_add(OBJECT_SPRITE_Z_ORDER);
  arrow->task = NULL;
  curf->object_header = g_list_append(curf->object_header, arrow);
  return (arrow);
}

static void init_object_info(int id, int type, gchar *ident_name, gchar *unident_name, int atk, int ac)
{
  Object_info[id].type = type;
  g_snprintf(Object_info[id].ident_name, OBJECT_NAME_MAX_LENGTH, "%s", ident_name);
  if (unident_name != NULL) {
    Object_info[id].ident = FALSE;
    g_snprintf(Object_info[id].unident_name, OBJECT_NAME_MAX_LENGTH, "%s", unident_name);
  } else {
    Object_info[id].ident = TRUE;
    g_snprintf(Object_info[id].unident_name, OBJECT_NAME_MAX_LENGTH, "NULL");
  };
  Object_info[id].atk = atk;
  Object_info[id].ac = ac;
}

void set_object_ident(struct _object *object, gboolean ident)
{
  Object_info[object->id].ident = ident;
}

gboolean object_type(struct _object *object)
{
  return Object_info[object->id].type;
}

gchar *object_name(struct _object *object)
{
  if (Object_info[object->id].ident)
    return Object_info[object->id].ident_name;
  return Object_info[object->id].unident_name;
}

gboolean object_ident(struct _object *object)
{
  return Object_info[object->id].ident;
}

int object_atk(struct _object *object)
{
  int atk;
  int plus;
  atk = Object_info[object->id].atk;
  plus = object->plus;
  return (atk + plus);
}

int object_ac(struct _object *object)
{
  int ac;
  int plus;
  ac = Object_info[object->id].ac;
  plus = object->plus;
  return (ac + plus);
}

gboolean object_cursed(struct _object *object)
{
  if (object->plus < 0) return TRUE;
  return FALSE;
};

void object_init()
{
  weapon_init();
  shield_init();
  arrow_init();
  potion_init();
  scroll_init();
  food_init();
  extra_init();
}

void object_free()
{
  int i;
  for (i = 0; i < OBJECT_TYPE_MAX; i++) {
    SDL_FreeSurface(surface_object_on_floor[i]);
  };
}

static int randomize_order(const void *a0, const void *a1)
{
  return g_random_int_range(-1, 2);
}

static void weapon_init()
{
  SDL_Surface *s;
  SDL_Rect r;
  s = video_bmp_load("bmp/weapon_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_WEAPON] = video_crop(s, &r);
  SDL_FreeSurface(s);
  init_object_info(OBJECT_ID_NOWEAPON, OBJECT_TYPE_WEAPON, "no weapon", NULL, 0, 0);
  init_object_info(OBJECT_ID_DAGGER, OBJECT_TYPE_WEAPON, "dagger", NULL, 1, 0);
  init_object_info(OBJECT_ID_MACE, OBJECT_TYPE_WEAPON, "mace", NULL, 2, 0);
  init_object_info(OBJECT_ID_LONGSWORD, OBJECT_TYPE_WEAPON, "longsword", NULL, 4, 0);
  init_object_info(OBJECT_ID_TWOHSWORD, OBJECT_TYPE_WEAPON, "twohsword", NULL, 6, 0);
}

static void shield_init()
{
  SDL_Surface *s;
  SDL_Rect r;
  s = video_bmp_load("bmp/shield_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_SHIELD] = video_crop(s, &r);
  SDL_FreeSurface(s);
  init_object_info(OBJECT_ID_NOSHIELD, OBJECT_TYPE_SHIELD, "no shield", NULL, 0, 0);
  init_object_info(OBJECT_ID_SMALL_SHIELD, OBJECT_TYPE_SHIELD, "small shield", NULL, 0, 2);
  init_object_info(OBJECT_ID_ORCISH_SHIELD, OBJECT_TYPE_SHIELD, "orcish shield", NULL, 0, 3);
  init_object_info(OBJECT_ID_ELVEN_SHIELD, OBJECT_TYPE_SHIELD, "elven shield", NULL, 0, 4);
  init_object_info(OBJECT_ID_LARGE_SHIELD, OBJECT_TYPE_SHIELD, "large shield", NULL, 0, 5);
  init_object_info(OBJECT_ID_DWARVISH_SHIELD, OBJECT_TYPE_SHIELD, "dwarvish shield", NULL, 0, 6);
}

static void arrow_init()
{
  SDL_Surface *s;
  SDL_Rect r;
  s = video_bmp_load("bmp/arrow_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_ARROW] = video_crop(s, &r);
  SDL_FreeSurface(s);
  init_object_info(OBJECT_ID_NOARROW, OBJECT_TYPE_ARROW, "no arrow", NULL, 0, 0);
  init_object_info(OBJECT_ID_WOOD_ARROW, OBJECT_TYPE_ARROW, "wooden arrow", NULL, 4, 0);
  init_object_info(OBJECT_ID_IRON_ARROW, OBJECT_TYPE_ARROW, "iron arrow", NULL, 12, 0);
}

static void potion_init()
{
  enum {COUNT_POTION = 5};
  SDL_Surface *s;
  SDL_Rect r;
  gchar buf[COUNT_POTION][OBJECT_NAME_MAX_LENGTH];
  int i;
  int table[COUNT_POTION];
  s = video_bmp_load("bmp/potion_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_POTION] = video_crop(s, &r);
  SDL_FreeSurface(s);
  g_snprintf(buf[0], OBJECT_NAME_MAX_LENGTH, "red potion");
  g_snprintf(buf[1], OBJECT_NAME_MAX_LENGTH, "blue potion");
  g_snprintf(buf[2], OBJECT_NAME_MAX_LENGTH, "yellow potion");
  g_snprintf(buf[3], OBJECT_NAME_MAX_LENGTH, "black potion");
  g_snprintf(buf[4], OBJECT_NAME_MAX_LENGTH, "white potion");
  for (i = 0; i < COUNT_POTION; i++) table[i] = i;
  qsort(table, COUNT_POTION, sizeof(int), randomize_order);
  init_object_info(OBJECT_ID_HEAL_POTION, OBJECT_TYPE_POTION, "healing potion", buf[table[0]], 0, 0);
  init_object_info(OBJECT_ID_INCSTR_POTION, OBJECT_TYPE_POTION,  "increase strength potion",  buf[table[1]], 0, 0);
  init_object_info(OBJECT_ID_RESSTR_POTION, OBJECT_TYPE_POTION, "restore strength potion", buf[table[2]], 0, 0);
  init_object_info(OBJECT_ID_POISON_POTION, OBJECT_TYPE_POTION, "poison potion",  buf[table[3]] ,0, 0);
  init_object_info(OBJECT_ID_RAISELEV_POTION, OBJECT_TYPE_POTION, "raise level potion",  buf[table[4]] ,0, 0);
}

static void scroll_init()
{
  enum {COUNT_SCROLL = 8};
  SDL_Surface *s;
  SDL_Rect r;
  gchar buf[COUNT_SCROLL][OBJECT_NAME_MAX_LENGTH];
  int i;
  int table[COUNT_SCROLL];
  s = video_bmp_load("bmp/scroll_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_SCROLL] = video_crop(s, &r);
  SDL_FreeSurface(s);
  g_snprintf(buf[0], OBJECT_NAME_MAX_LENGTH, "green scroll");
  g_snprintf(buf[1], OBJECT_NAME_MAX_LENGTH, "yellow scroll");
  g_snprintf(buf[2], OBJECT_NAME_MAX_LENGTH, "red scroll");
  g_snprintf(buf[3], OBJECT_NAME_MAX_LENGTH, "white scroll");
  g_snprintf(buf[4], OBJECT_NAME_MAX_LENGTH, "blue scroll");
  g_snprintf(buf[5], OBJECT_NAME_MAX_LENGTH, "black scroll");
  g_snprintf(buf[6], OBJECT_NAME_MAX_LENGTH, "purple scroll");
  g_snprintf(buf[7], OBJECT_NAME_MAX_LENGTH, "gray scroll");
  for (i = 0; i < COUNT_SCROLL; i++) table[i] = i;
  qsort(table, COUNT_SCROLL, sizeof(int), randomize_order);
  init_object_info(OBJECT_ID_IDENT_SCROLL, OBJECT_TYPE_SCROLL, "identify scroll", buf[table[0]], 0, 0);
  init_object_info(OBJECT_ID_RMCURSE_SCROLL, OBJECT_TYPE_SCROLL, "remove curse scroll", buf[table[1]], 0, 0);
  init_object_info(OBJECT_ID_ENCWEAPON_SCROLL, OBJECT_TYPE_SCROLL, "enchant weapon scroll", buf[table[2]], 0, 0);
  init_object_info(OBJECT_ID_ENCSHIELD_SCROLL, OBJECT_TYPE_SCROLL, "enchant shield scroll", buf[table[3]], 0, 0);
  init_object_info(OBJECT_ID_EXPLOSION_SCROLL, OBJECT_TYPE_SCROLL, "explosion scroll", buf[table[4]], 0, 0);
  init_object_info(OBJECT_ID_MONDETECT_SCROLL, OBJECT_TYPE_SCROLL, "monster detect scroll", buf[table[5]], 0, 0);
  init_object_info(OBJECT_ID_OBJDETECT_SCROLL, OBJECT_TYPE_SCROLL, "object detect scroll", buf[table[6]], 0, 0);
  init_object_info(OBJECT_ID_MAP_SCROLL, OBJECT_TYPE_SCROLL, "map scroll", buf[table[7]], 0, 0);
}

static void food_init()
{
  SDL_Surface *s;
  SDL_Rect r;
  s = video_bmp_load("bmp/food_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_FOOD] = video_crop(s, &r);
  SDL_FreeSurface(s);
  init_object_info(OBJECT_ID_RATION, OBJECT_TYPE_FOOD, "ration", NULL, 0, 0);
  init_object_info(OBJECT_ID_SLIMEMOLD, OBJECT_TYPE_FOOD, "slimemold", NULL, 0, 0);
}

static void extra_init()
{
  SDL_Surface *s;
  SDL_Rect r;
  s = video_bmp_load("bmp/extra_on_floor.bmp");
  r.x = 0; r.y = 0;
  r.w = SPRITE_W; r.h = SPRITE_H;
  surface_object_on_floor[OBJECT_TYPE_EXTRA] = video_crop(s, &r);
  SDL_FreeSurface(s);
  init_object_info(OBJECT_ID_AMUYENDOR, OBJECT_TYPE_EXTRA, "amulet of yendor", "amulet of yendor", 0, 0);
}
 
