#include <stdlib.h>
#include <glib.h>
#include <SDL.h>
#include "racanhack.h"
#include "video.h"
#include "task.h"
#include "cave.h"
#include "player.h"

static GList *sprite_header[Z_ORDER];

void SDL_init()
{
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
    g_printerr("Unable to init SDL: %s\n", SDL_GetError());
    exit(1);
  };
  atexit(SDL_Quit);
  screen = SDL_SetVideoMode(DISPLAY_W, DISPLAY_H, DISPLAY_BPP, SDL_SWSURFACE | SDL_DOUBLEBUF);
  if (!screen) {
    g_printerr("Unable to set %dx%d video: %s\n", DISPLAY_W, DISPLAY_H, SDL_GetError());
    exit(1);
  };
  transparent = SDL_MapRGB(screen->format, 240, 0, 240);
}

void video_sprite_init()
{
  int i;
  for (i = 0; i < Z_ORDER; i++) {
    sprite_header[i] = NULL;
  };
}

void video_sprite_free()
{
  int i;
  for (i = 0; i < Z_ORDER; i++) {
    g_list_foreach(sprite_header[i], (GFunc)my_free, NULL);
    g_list_free(sprite_header[i]);
  };
}

struct _sprite *video_sprite_add(int z_order)
{
  struct _sprite *sprite;
  g_assert((z_order < Z_ORDER) && (z_order >= 0));
  sprite = my_new(struct _sprite, 1);
  sprite->x = 0;
  sprite->y = 0;
  sprite->abs = FALSE;
  sprite->alive = TRUE;
  sprite->surface = NULL;
  sprite_header[z_order] = g_list_append(sprite_header[z_order], sprite);
  return(sprite);
}

SDL_Surface *video_bmp_load(char *file)
{
  SDL_Surface *temp, *surface;
  temp = SDL_LoadBMP(file);
  if (!(temp)) {
    g_printerr("Couldn't load %s: %s\n", file, SDL_GetError());
    exit(1);
  };
  surface = SDL_DisplayFormat(temp);
  if (!(surface)) {
    g_printerr("Couldn't DisplayFormat %s: %s\n", file, SDL_GetError());
    exit(1);
  };
  return surface;
}

SDL_Surface* video_crop(SDL_Surface *s, SDL_Rect *r)
{
  SDL_Surface *temp, *surface;
  temp = SDL_CreateRGBSurface(SDL_SWSURFACE, r->w, r->h, screen->format->BitsPerPixel, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
  if (!(temp)) {
    g_printerr("failed CreateRGBSurface : %s\n", SDL_GetError());
    exit(1);
  };
  SDL_BlitSurface(s, r, temp, NULL);
  SDL_SetColorKey(temp, SDL_SRCCOLORKEY | SDL_RLEACCEL, transparent);
  surface = SDL_DisplayFormat(temp);
  if (!(surface)) {
    g_printerr("Couldn't DisplayFormat : %s\n", SDL_GetError());
    exit(1);
  };
  return surface;
}

SDL_Surface* video_crop_alpha(SDL_Surface *s, SDL_Rect *r, int alpha)
{
  SDL_Surface *temp, *surface;
  temp = SDL_CreateRGBSurface(SDL_SWSURFACE, r->w, r->h, screen->format->BitsPerPixel, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
  if (!(temp)) {
    g_printerr("failed CreateRGBSurface : %s\n", SDL_GetError());
    exit(1);
  };
  SDL_BlitSurface(s, r, temp, NULL);
  SDL_SetColorKey(temp, SDL_SRCCOLORKEY | SDL_RLEACCEL, transparent);
  SDL_SetAlpha(temp, SDL_SRCALPHA | SDL_RLEACCEL, alpha);
  surface = SDL_DisplayFormat(temp);
  if (!(surface)) {
    g_printerr("Couldn't DisplayFormat : %s\n", SDL_GetError());
    exit(1);
  };
  return surface;
}

void video_display()
{
  int i;
  GList *li, *next;
  struct _sprite *sprite;
  int x, y;
  SDL_Rect r;
  SDL_FillRect(screen, NULL,0x000000);
  for (i = 0; i < Z_ORDER; i++) {
    for (li = g_list_first(sprite_header[i]); li != NULL; li = next) {
      next = g_list_next(li);
      sprite = (struct _sprite *)li->data;
      if (sprite->alive == FALSE) {
	sprite_header[i] = g_list_remove(sprite_header[i], sprite);
	my_free(sprite);
	continue;
      };
      if (sprite->surface == NULL) continue;
      if (sprite->abs == FALSE) {
	x = sprite->x - Player.sprite->x + (DISPLAY_W / 2);
	y = sprite->y - Player.sprite->y + (DISPLAY_H / 2);
      } else {
	x = sprite->x;
	y = sprite->y;
      };
      r.x = x;
      r.y = y;
      r.w = sprite->surface->w;
      r.h = sprite->surface->h;
      SDL_BlitSurface(sprite->surface, NULL, screen, &r);
    };
  };
  SDL_Flip(screen);
}
